// Before the definition of YYLTYPE and YYSTYPE.
%code requires
{
# include <stdio.h>
# include <stdlib.h>
}

// After the definition of YYLTYPE and YYSTYPE.
%code provides
{
#define YY_DECL						\
  enum yytokentype yylex(YYSTYPE* lval, YYLTYPE* lloc)

  YY_DECL;

  // YYLTYPE: type for locations.
  // errors: a user defined argument.
  // msg: the error message, generated by the parser.
  void yyerror (YYLTYPE* loc, int* errors, const char* msg);
}

 // Instrument the parser for tracing (see yydebug below).
%debug
 // Better error messages (with expected/unexpected tokens).
%error-verbose
 // Accept no conflicts.
%expect 0
 // Enable location tracking.
%locations
 // User defined argument for yyparse (and yyerror).
%parse-param { int* errors }
 // Make yylval and yylloc local variables instead of globals.
%pure-parser

 // Be ready to store semantical value as ints.
%union
{
  int ival;
};

 // Define the tokens, and their type.
%token PLUS "+" STAR "*" MINUS "-" SLASH "/"
  LPAREN "(" RPAREN ")"
  EOL  "end of line"
%token <ival> NUM "number"
 // Non-terminal type for semantical values.
%type <ival> exp line

%left "+" "-"
%left "*" "/"

%printer { fprintf (stderr, "%d", $$); } <ival>

%%

 /*----------.
 | Grammar.  |
 `----------*/

lines:
  /* empty */
| lines line    
;

line:
  "end of line"       {}
| exp "end of line"   { printf ("=> %d\n", $1); }
| error "end of line" { yyerrok; }
;

exp:
  exp "+" exp { $$ = $1 + $3; }
| exp "*" exp { $$ = $1 * $3; }
| exp "-" exp { $$ = $1 - $3; }
| exp "/" exp {
                if ($3)
                  $$ = $1 / $3; 
		else
		{
		  yyerror(&@3, errors, "invalid denominator");
		  YYERROR;
		}
              }
| "(" exp ")" { $$ = $2; }
| "+" exp     { $$ = $2; }
| "-" exp     { $$ = - $2; }
| "number"    { $$ = $1; }
;

%%

void
yyerror (YYLTYPE* loc, int* my_errors, const char* msg)
{
  YY_LOCATION_PRINT(stderr, *loc);
  fprintf (stderr, ": %s\n", msg);
  ++*my_errors;
}

int
main()
{
  yydebug = !!getenv("PARSER");
  int errors = 0;
  yyparse(&errors);
  return !!errors;
}
